from ...pyhwException import OSUnsupportedException


class UptimeDetect:
    """
    Class for system uptime detection.
    """
    def __init__(self, os):
        self.OS = os

    def getUptime(self):
        """
        Detects the system uptime.
        :return: dataclass UptimeInfo, direct attr: uptime
        """
        if self.OS == "linux":
            from .linux import UptimeDetectLinux
            return UptimeDetectLinux().getUptimeInfo()
        elif self.OS == "macos":
            from .macos import UptimeDetectMacOS
            return UptimeDetectMacOS().getUptimeInfo()
        elif self.OS == "freebsd":
            from .bsd import UptimeDetectBSD
            return UptimeDetectBSD().getUptimeInfo()
        elif self.OS == "windows":
            from .windows import UptimeDetectWindows
            return UptimeDetectWindows().getUptimeInfo()
        else:
            raise OSUnsupportedException("Unsupported operating system")



from ...pyhwException import BackendException
from .memoryInfo import MemoryInfo
import subprocess
import json


class MemoryDetectWindows:
    def __init__(self):
        self._memoryInfo = MemoryInfo()

    def getMemoryInfo(self):
        self._getMemory()
        self._memoryInfo.memory = f"{self._memoryInfo.used} MiB / {self._memoryInfo.total} MiB"
        return self._memoryInfo

    def _getMemory(self):
        COMMAND = 'Get-CimInstance -ClassName Win32_OperatingSystem | Select-Object FreePhysicalMemory,TotalVisibleMemorySize | ConvertTo-JSON'
        try:
            result = subprocess.run(["powershell", "-NoProfile", "-Command", COMMAND], capture_output=True, text=True)
        except subprocess.SubprocessError:
            raise BackendException("Failed to get memory information on Windows.")

        res = json.loads(result.stdout)
        free_memory = int(res['FreePhysicalMemory']) / 1024
        total_memory = int(res['TotalVisibleMemorySize']) / 1024
        used_memory = total_memory - free_memory
        self._memoryInfo.total = round(total_memory, 2)
        self._memoryInfo.available = round(free_memory, 2)
        self._memoryInfo.used = round(used_memory, 2)










import subprocess
from .nicInfo import NICInfo
from ...pyhwUtil import getArch
from ...pyhwException import BackendException
import pypci
import os


class NICDetectWindows:
    def __init__(self):
        self._nicInfo = NICInfo()

    def getNICInfo(self):
        self._getNICInfo()
        self._sortNICList()
        return self._nicInfo

    def _getNICInfo(self):
        nic_devices = pypci.PCI().FindAllNIC()
        if len(nic_devices) == 0:
            self.__handleNonePciDevices()
        else:
            for device in nic_devices:
                if device.subsystem_device_name != "":
                    device_name = f"{device.vendor_name} {device.device_name} ({device.subsystem_device_name})"
                else:
                    device_name = f"{device.vendor_name} {device.device_name}"
                self._nicInfo.nics.append(self._nicNameClean(device_name))
                self._nicInfo.number += 1

    def __handleNonePciDevices(self):
        # need to update
        self._nicInfo.nics.append("Not found")
        self._nicInfo.number = 1

    @staticmethod
    def _nicNameClean(nic_name: str):
        nic_name_clean = nic_name.replace("Corporation ", "")
        return nic_name_clean

    def _sortNICList(self):
        return self._nicInfo.nics.sort()





"""
    In dev.
"""
from ...pyhwUtil import getDocker
from .shellInfo import ShellInfo
import os
import subprocess


class ShellDetectUnix:
    def __init__(self):
        self.__shellInfo = ShellInfo()

    def getShellInfo(self):
        if getDocker():
            self.__getShellDocker()
        else:
            self.__getShell()
        self.__getVersion()
        self.__shellInfo.info = self.__shellInfo.shell + " " + self.__shellInfo.version
        return self.__shellInfo

    def __getShell(self):
        # Get the default shell, not the current shell.
        shell_env = os.getenv("SHELL", "")
        if shell_env != "":
            self.__shellInfo.shell = shell_env.split("/")[-1]
            self.__shellInfo.path = shell_env

    def __getShellDocker(self):
        try:
            with open("/etc/passwd", "r") as f:
                for line in f:
                    if line.startswith("root:"):
                        root_shell = line.strip().split(":")[-1]
                        break
        except:
            root_shell = ""
        if root_shell != "":
            self.__shellInfo.shell = root_shell.split("/")[-1]
            self.__shellInfo.path = root_shell

    def __getVersion(self):
        shell = self.__shellInfo.shell
        if shell != "":
            if shell in ["sh", "ash", "dash", "es"]:
                pass
            elif shell == "bash":
                try:
                    result = subprocess.run(["bash", "-c", "echo $BASH_VERSION"], capture_output=True, text=True)
                    self.__shellInfo.version = result.stdout.strip().split("(")[0]
                except subprocess.SubprocessError:
                    pass
            elif shell == "zsh":
                try:
                    result = subprocess.run(["zsh", "-c", "echo $ZSH_VERSION"], capture_output=True, text=True)
                    self.__shellInfo.version = result.stdout.strip()
                except subprocess.SubprocessError:
                    pass










from dataclasses import dataclass


@dataclass
class ColorSet:
    COLOR_MODE_RESET = "0"
    COLOR_MODE_BOLD = "1"
    COLOR_MODE_DIM = "2"
    COLOR_MODE_ITALIC = "3"
    COLOR_MODE_UNDERLINE = "4"
    COLOR_MODE_BLINK = "5"
    COLOR_MODE_INVERSE = "7"
    COLOR_MODE_HIDDEN = "8"
    COLOR_MODE_STRIKETHROUGH = "9"

    COLOR_FG_BLACK = "30"
    COLOR_FG_RED = "31"
    COLOR_FG_GREEN = "32"
    COLOR_FG_YELLOW = "33"
    COLOR_FG_BLUE = "34"
    COLOR_FG_MAGENTA = "35"
    COLOR_FG_CYAN = "36"
    COLOR_FG_WHITE = "37"
    COLOR_FG_DEFAULT = "39"

    COLOR_FG_LIGHT_BLACK = "90"
    COLOR_FG_LIGHT_RED = "91"
    COLOR_FG_LIGHT_GREEN = "92"
    COLOR_FG_LIGHT_YELLOW = "93"
    COLOR_FG_LIGHT_BLUE = "94"
    COLOR_FG_LIGHT_MAGENTA = "95"
    COLOR_FG_LIGHT_CYAN = "96"
    COLOR_FG_LIGHT_WHITE = "97"

    COLOR_BG_BLACK = "40"
    COLOR_BG_RED = "41"
    COLOR_BG_GREEN = "42"
    COLOR_BG_YELLOW = "43"
    COLOR_BG_BLUE = "44"
    COLOR_BG_MAGENTA = "45"
    COLOR_BG_CYAN = "46"
    COLOR_BG_WHITE = "47"
    COLOR_BG_DEFAULT = "49"

    COLOR_BG_LIGHT_BLACK = "100"
    COLOR_BG_LIGHT_RED = "101"
    COLOR_BG_LIGHT_GREEN = "102"
    COLOR_BG_LIGHT_YELLOW = "103"
    COLOR_BG_LIGHT_BLUE = "104"
    COLOR_BG_LIGHT_MAGENTA = "105"
    COLOR_BG_LIGHT_CYAN = "106"
    COLOR_BG_LIGHT_WHITE = "107"

    COLOR_FG_256 = "38;5;"
    COLOR_BG_256 = "48;5;"

    COLOR_FG_RGB = "38;2;"
    COLOR_BG_RGB = "48;2;"



from ...pyhwException import LogoNotFoundException
from pathlib import Path
import os


class Logo:
    def __init__(self, logo_os):
        self.__logo_os = logo_os
        self.__logo_ascii = ""

    def getLogoContent(self):
        self.__loadLogoAscii()
        return self.__logo_ascii

    def __loadLogoAscii(self):
        try:
            file_path = os.path.join(Path(__file__).parent, f"ascii/{self.__logo_os}.pyhw")
            with open(file_path, "r") as f:
                self.__logo_ascii = f.read()
        except FileNotFoundError:
            raise LogoNotFoundException(f"Logo for {self.__logo_os} not found.")




from .memoryInfo import MemoryInfo
from ...pyhwUtil import sysctlGetInt
import subprocess


class MemoryDetectMacOS:
    def __init__(self):
        self.__memoryInfo = MemoryInfo()
        self.__hw_pagesize = 0
        self.__mem_total = 0
        self.__active_count = 0
        self.__inactive_count = 0
        self.__speculative_count = 0
        self.__wire_count = 0
        self.__compressor_page_count = 0
        self.__purgeable_count = 0
        self.__external_page_count = 0
        self.__mem_used = 0

    def getMemoryInfo(self):
        self.__getMemory()
        self.__memoryInfo.total = self.__mem_total / 1024 / 1024  # Convert bytes to MiB
        self.__memoryInfo.used = self.__mem_used / 1024 / 1024  # Convert bytes to MiB
        self.__memoryInfo.memory = f"{round(self.__memoryInfo.used, 2)} MiB / {round(self.__memoryInfo.total, 2)} MiB"
        return self.__memoryInfo

    def __getMemory(self):
        self.__hw_pagesize = sysctlGetInt("hw.pagesize")
        self.__mem_total = sysctlGetInt("hw.memsize")
        self.__getVMStat()
        self.__mem_used = self.__hw_pagesize * (self.__active_count + self.__inactive_count + self.__speculative_count +
                                                self.__wire_count + self.__compressor_page_count -
                                                self.__purgeable_count - self.__external_page_count)

    def __getVMStat(self):
        try:
            result = subprocess.run(["vm_stat"], capture_output=True, text=True)
            vm_stats = result.stdout.split("\n")
        except subprocess.SubprocessError:
            vm_stats = []
        for vm_stat in vm_stats:
            if vm_stat.startswith("Pages active"):
                self.__active_count = int(vm_stat.split(":")[1].strip().split(".")[0])
            elif vm_stat.startswith("Pages inactive"):
                self.__inactive_count = int(vm_stat.split(":")[1].strip().split(".")[0])
            elif vm_stat.startswith("Pages speculative"):
                self.__speculative_count = int(vm_stat.split(":")[1].strip().split(".")[0])
            elif vm_stat.startswith("Pages wired down"):
                self.__wire_count = int(vm_stat.split(":")[1].strip().split(".")[0])
            elif vm_stat.startswith("Pages occupied by compressor"):  # compressor_page_count
                self.__compressor_page_count = int(vm_stat.split(":")[1].strip().split(".")[0])
            elif vm_stat.startswith("Pages purgeable"):
                self.__purgeable_count = int(vm_stat.split(":")[1].strip().split(".")[0])
            # miss external_page_count here
            elif vm_stat.startswith("File-backed pages:"):
                self.__external_page_count = int(vm_stat.split(":")[1].strip().split(".")[0])



from .pyhwUtil import getOS, getArch, getDocker, createDataString, selectOSLogo
from .sysctlUtil import sysctlGetString, sysctlGetInt

__all__ = ["getOS", "getArch", "getDocker", "createDataString", "selectOSLogo", "sysctlGetString", "sysctlGetInt"]



from dataclasses import dataclass


@dataclass
class Data:
    title = "Default Value"
    underline = "Default Value"
    OS = "Default Value"
    Host = "Default Value"
    Kernel = "Default Value"
    Uptime = "Default Value"
    # Packages = ""
    Shell = "Default Value"
    # Resolution = ""
    # DE = ""
    # WM = ""
    # WM_Theme = ""
    # Terminal = ""
    CPU = "Default Value"
    GPU = ["Default Value"]
    Memory = "Default Value"
    NIC = ["Default Value"]
    NPU = ["Default Value"]



"""
 In dev.
"""
from .titleInfo import TitleInfo
import getpass
import platform


class TitleDetectWindows:
    def __init__(self):
        self.__titleInfo = TitleInfo()

    def getTitle(self):
        self.__titleInfo.username = getpass.getuser()
        self.__titleInfo.hostname = platform.node()
        self.__titleInfo.title = f"{self.__titleInfo.username}@{self.__titleInfo.hostname}"
        return self.__titleInfo



from ...pyhwException import OSUnsupportedException


class MemoryDetect:
    def __init__(self, os):
        self.OS = os

    def getMemoryInfo(self):
        if self.OS == "linux":
            from .linux import MemoryDetectLinux
            return MemoryDetectLinux().getMemoryInfo()
        elif self.OS == "macos":
            from .macos import MemoryDetectMacOS
            return MemoryDetectMacOS().getMemoryInfo()
        elif self.OS == "freebsd":
            from .bsd import MemoryDetectBSD
            return MemoryDetectBSD().getMemoryInfo()
        elif self.OS == "windows":
            from .windows import MemoryDetectWindows
            return MemoryDetectWindows().getMemoryInfo()
        else:
            raise OSUnsupportedException("Unsupported operating system")



"""
    In dev.
"""
from ...pyhwUtil import getArch
from ...pyhwException import BackendException
from .hostInfo import HostInfo
import winreg


class HostDetectWindows:
    def __init__(self):
        self._hostInfo = HostInfo()
        self._arch = getArch()

    def getHostInfo(self):
        self._getModel()
        return self._hostInfo

    def _getModel(self):
        try:
            key = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, r"HARDWARE\DESCRIPTION\System\BIOS")
            manufacturer, _ = winreg.QueryValueEx(key, "SystemManufacturer")
            name, _ = winreg.QueryValueEx(key, "SystemProductName")
            version, _ = winreg.QueryValueEx(key, "SystemVersion")
            winreg.CloseKey(key)
        except FileNotFoundError:
            raise BackendException("Unable to retrieve system information.")
        self._hostInfo.name = name
        self._hostInfo.version = version
        self._hostInfo.vendor = manufacturer
        self.__handleVM()
        self._hostInfo.model = self._hostInfo.name + " " + self._hostInfo.version

    def __handleVM(self):
        if self._hostInfo.name.startswith("VMware"):
            self._hostInfo.version = ""



"""
    In dev.
"""
from .kernelInfo import KernelInfo
from ...pyhwException import BackendException
import platform
import winreg


class KernelDetectWindows:
    def __init__(self):
        self.__kernelInfo = KernelInfo()

    def getKernelInfo(self):
        version = platform.version()
        machine = platform.machine()
        display = self.__get_windows_version()
        if display != "":
            self.__kernelInfo.kernel = f"{version} ({display}) {machine}"
        else:
            self.__kernelInfo.kernel = f"{version} {machine}"
        return self.__kernelInfo

    @staticmethod
    def __get_windows_version():
        try:
            key = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, r"SOFTWARE\Microsoft\Windows NT\CurrentVersion")
            display_version, _ = winreg.QueryValueEx(key, "DisplayVersion")
            winreg.CloseKey(key)
            return str(display_version)
        except:
            try:
                key = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, r"SOFTWARE\Microsoft\Windows NT\CurrentVersion")
                release_id, _ = winreg.QueryValueEx(key, "ReleaseId")
                winreg.CloseKey(key)
                return str(release_id)
            except:
                return ""



from .npuInfo import NPUInfo
import pypci


class NPUDetectWindows:
    def __init__(self):
        self._npuInfo = NPUInfo()

    def getNPUInfo(self):
        self._getNPUInfo()
        self._sortNPUList()
        return self._npuInfo

    def _getNPUInfo(self):
        npu_devices = pypci.PCI().FindAllNPU()
        if len(npu_devices) == 0:
            self._handleNonePciDevices()
        else:
            for device in npu_devices:
                if device.subsystem_device_name != "":
                    device_name = f"{device.vendor_name} {device.device_name} ({device.subsystem_device_name})"
                else:
                    device_name = f"{device.vendor_name} {device.device_name}"
                self._npuInfo.npus.append(self._npuNameClean(device_name))
                self._npuInfo.number += 1

    def _handleNonePciDevices(self):
        # Place Holder for unknown NPU
        self._npuInfo.number = 1
        self._npuInfo.npus.append("Not found")

    @staticmethod
    def _npuNameClean(npu_name: str):
        npu_name_clean = npu_name.replace("Corporation ", "")
        return npu_name_clean

    def _sortNPUList(self):
        self._npuInfo.npus.sort()





from .linux import HostDetectLinux


class HostDetectBSD(HostDetectLinux):
    def __init__(self):
        HostDetectLinux.__init__(self)

    def getHostInfo(self):
        self._hostInfo.name = f"General {self._arch} FreeBSD Host"
        self._hostInfo.version = ""
        self._hostInfo.model = self._hostInfo.name + " " + self._hostInfo.version
        return self._hostInfo




import platform
from ..backend import Data
import os
from dataclasses import dataclass


def getOS():
    """
    Get the os type in lower case.
    :return: str, os type, value in [windows, linux, macos, freebsd, unknown].
    """
    system = platform.system()
    if system == "Windows":
        return "windows"
    elif system == "Linux":
        return "linux"
    elif system == "Darwin":
        return "macos"
    elif system == "FreeBSD":
        return "freebsd"
    else:
        return "unknown"


def getArch():
    """
    Get the machine architecture.
    :return: str, value in [x86_64, x86, aarch64, arm32, riscv64].
    """
    arch = platform.machine()
    if arch == "x86_64" or arch == "AMD64" or arch == "amd64":
        return "x86_64"
    elif arch == "i386" or arch == "i686" or arch == "x86":
        return "x86"
    elif arch == "aarch64" or arch == "arm64":
        return "aarch64"
    elif arch.find("arm") != -1:
        return "arm32"
    elif arch == "riscv64":
        return "riscv64"
    else:
        return "unknown"


def getDocker() -> bool:
    """
    Check if the current environment is running in Docker.
    :return: bool, True if running in Docker, False otherwise.
    """
    return os.path.exists("/.dockerenv")


class DataStringProcessor:
    def __init__(self, data: Data):
        self.data = data
        self.columns = self.__getENV()

    @staticmethod
    def __getENV() -> int:
        if getOS() == "linux":
            try:
                _, columns_str = os.popen('stty size', 'r').read().split()
                columns = int(columns_str)
            except:
                columns = 80  # default terminal size is 80 columns
        else:
            # macOS default terminal size is 80 columns
            columns = 80
        return columns

    def __dropLongString(self, string: str) -> str:
        """
        Drop the string if it's too long to fit in the terminal.
        :param string: str, the input string.
        :return: str, the shortened string, do not include newline char.
        """
        if len(string) >= self.columns:
            return f"{string[:self.columns-1]}"
        else:
            return f"{string}"

    def getTitle(self) -> str:
        return f" {self.data.title}\n"

    def getLine(self) -> str:
        return f" {'-'*len(self.data.title)}\n"

    def getOS(self) -> str:
        os_str = f" OS: {self.data.OS}"
        return f"{self.__dropLongString(os_str)}\n"

    def getHost(self) -> str:
        host_str = f" Host: {self.data.Host}"
        return f"{self.__dropLongString(host_str)}\n"

    def getKernel(self) -> str:
        kernel_str = f" Kernel: {self.data.Kernel}"
        return f"{self.__dropLongString(kernel_str)}\n"

    def getUptime(self) -> str:
        uptime_str = f" Uptime: {self.data.Uptime}"
        return f"{self.__dropLongString(uptime_str)}\n"

    def getShell(self) -> str:
        shell_str = f" Shell: {self.data.Shell}"
        return f"{self.__dropLongString(shell_str)}\n"

    def getCPU(self) -> str:
        cpu_str = f" CPU: {self.data.CPU}"
        return f"{self.__dropLongString(cpu_str)}\n"

    def getGPU(self) -> str:
        ret_str = ""
        for gpu in self.data.GPU:
            gpu_str = f" GPU: {gpu}"
            ret_str += f"{self.__dropLongString(gpu_str)}\n"
        return ret_str

    def getMemory(self) -> str:
        memory_str = f" Memory: {self.data.Memory}"
        return f"{self.__dropLongString(memory_str)}\n"

    def getNIC(self) -> str:
        ret_str = ""
        for nic in self.data.NIC:
            nic_str = f" NIC: {nic}"
            ret_str += f"{self.__dropLongString(nic_str)}\n"
        return ret_str

    def getNPU(self) -> str:
        ret_str = ""
        for npu in self.data.NPU:
            npu_str = f" NPU: {npu}"
            ret_str += f"{self.__dropLongString(npu_str)}\n"
        return ret_str


def createDataString(data: Data):
    data_string_processor = DataStringProcessor(data)
    data_string = ""
    data_string += data_string_processor.getTitle()
    data_string += data_string_processor.getLine()
    data_string += data_string_processor.getOS()
    data_string += data_string_processor.getHost()
    data_string += data_string_processor.getKernel()
    data_string += data_string_processor.getUptime()
    data_string += data_string_processor.getShell()
    data_string += data_string_processor.getCPU()
    data_string += data_string_processor.getGPU()
    data_string += data_string_processor.getNPU()
    data_string += data_string_processor.getMemory()
    data_string += data_string_processor.getNIC()
    return data_string


def createDataStringOld(data: Data):
    data_string = ""
    data_string += f" {data.title}\n"
    data_string += f" {'-'*len(data.title)}\n"
    data_string += f" OS: {data.OS}\n"
    data_string += f" Host: {data.Host}\n"
    data_string += f" Kernel: {data.Kernel}\n"
    data_string += f" Uptime: {data.Uptime}\n"
    data_string += f" Shell: {data.Shell}\n"
    data_string += f" CPU: {data.CPU}\n"
    for gpu in data.GPU:
        data_string += f" GPU: {gpu}\n"
    data_string += f" Memory: {data.Memory}\n"
    for nic in data.NIC:
        data_string += f" NIC: {nic}\n"
    for npu in data.NPU:
        data_string += f" NPU: {npu}\n"
    return data_string


@dataclass
class SupportedOS:
    ColorConfig = ["armbian", "arch", "alpine", "centos", "debian", "fedora", "macOS", "raspbian", "ubuntu", "freebsd",
                   "windows_old", "windows_10", "windows_11", "windows_2025"]
    AsciiLogo = ["armbian", "arch", "alpine", "centos", "debian", "fedora", "macOS", "raspbian", "ubuntu", "freebsd",
                 "windows_old", "windows_10", "windows_11", "windows_2025"]


def selectOSLogo(os_id: str):
    """
    Select the logo based on the os id and terminal size.
    :param os_id: str, os id.
    :return: str, logo id.
    """
    if getOS() == "macos":  # macOS does not have small logo
        return os_id
    if getOS() == "windows":    # windows does not have small logo
        if os_id in SupportedOS.ColorConfig and os_id in SupportedOS.AsciiLogo:
            return os_id
        else:
            return "windows_11"
    if os_id in SupportedOS.ColorConfig and os_id in SupportedOS.AsciiLogo:
        pass
    else:
        return "linux"
    try:
        rows_str, columns_str = os.popen('stty size', 'r').read().split()
        rows = int(rows_str)
        columns = int(columns_str)
    except:
        # Default terminal size is 24 rows and 80 columns
        rows = 24
        columns = 80
    if columns <= 80:
        if os_id in ["fedora", "ubuntu"]:
            return f"{os_id}_small"
        else:
            return os_id
    else:
        return os_id




from ...pyhwException import OSUnsupportedException


class OSDetect:
    def __init__(self, os):
        self.__OS = os

    def getOSInfo(self):
        if self.__OS == "linux":
            from .linux import OSDetectLinux
            return OSDetectLinux().getOSInfo()
        elif self.__OS == "macos":
            from .macos import OSDetectMacOS
            return OSDetectMacOS().getOSInfo()
        elif self.__OS == "freebsd":
            from .bsd import OSDetectBSD
            return OSDetectBSD().getOSInfo()
        elif self.__OS == "windows":
            from .windows import OSDetectWindows
            return OSDetectWindows().getOSInfo()
        else:
            raise OSUnsupportedException("Unsupported operating system")



from ...pyhwException import OSUnsupportedException


class KernelDetect:
    def __init__(self, os):
        self.OS = os

    def getKernelInfo(self):
        if self.OS in ["linux", "macos", "freebsd"]:
            from .unix import KernelDetectUnix
            return KernelDetectUnix().getKernelInfo()
        elif self.OS == "windows":
            from .windows import KernelDetectWindows
            return KernelDetectWindows().getKernelInfo()
        else:
            raise OSUnsupportedException("Unsupported operating system")



"""
    In dev.
"""
from .uptimeInfo import UptimeInfo


class UptimeDetectLinux:
    def __init__(self):
        self.__uptimeInfo = UptimeInfo()

    def getUptimeInfo(self):
        self.__getUptime()
        return self.__uptimeInfo

    def __getUptime(self):
        with open("/proc/uptime", "r") as f:
            uptime_seconds = float(f.readline().split(" ")[0])
            hours, remainder = divmod(uptime_seconds, 3600)
            minutes, seconds = divmod(remainder, 60)
            days, hours = divmod(hours, 24)
            days = int(days)
            hours = int(hours)
            minutes = int(minutes)
            seconds = int(seconds)
            if days == 0:
                if hours == 0:
                    self.__uptimeInfo.uptime = f"{minutes} mins {seconds} secs"
                else:
                    self.__uptimeInfo.uptime = f"{hours} hours {minutes} mins {seconds} secs"
            else:
                self.__uptimeInfo.uptime = f"{days} days {hours} hours {minutes} mins {seconds} secs"



from dataclasses import dataclass


@dataclass
class KernelInfo:
    """
    Data class for storing kernel information.
    """
    def __init__(self):
        """
        Initialize the data class, kernel attribute is required.
        """
        self.name = ""
        self.version = ""
        self.machine = ""
        self.kernel = ""



import subprocess


def __sysctlGet(key):
    """
    Get value from sysctl by key.
    :param key: str, sysctl key
    :return: value of sysctl key or "" if not found
    """
    try:
        output = subprocess.run(["sysctl", "-n", key], capture_output=True, text=True).stdout.strip()
    except subprocess.SubprocessError:
        output = ""
    return output


def sysctlGetString(key):
    """
    Get value from sysctl by key.
    :param key: str, sysctl key
    :return: value of sysctl key or "" if not found
    """
    return __sysctlGet(key)


def sysctlGetInt(key):
    """
    Get value from sysctl by key.
    :param key: str, sysctl key
    :return: value of sysctl key as int or None if not found
    """
    value = __sysctlGet(key)
    try:
        res = int(value)
    except Exception:
        res = None
    return res



from .gpuInfo import GPUInfo
from ...pyhwUtil import getArch
import json
import subprocess
import ctypes
from pathlib import Path


class GPUDetectMacOS:
    def __init__(self):
        self.__gpuInfo = GPUInfo()
        self.__arch = getArch()

    def getGPUInfo(self):
        if self.__arch == "aarch64":
            self.__getGPUAppleSilicon()
        else:   # Does not consider powerPC based Macs.
            self.__getGPUIntel()
        return self.__gpuInfo

    def __getGPUAppleSilicon(self):
        gpus = list()
        try:
            gpu_info_dict = json.loads(subprocess.check_output(["system_profiler", "SPDisplaysDataType", "-json"]))
            if 'SPDisplaysDataType' in gpu_info_dict:
                gpus = gpu_info_dict['SPDisplaysDataType']
                self.__gpuInfo.number = len(gpus)
            else:
                pass
        except Exception:
            return

        for gpu in gpus:
            self.__gpuInfo.gpus.append(f'{gpu.get("sppci_model")} ({gpu.get("sppci_cores")} cores) [SOC Integrated]')

    def __getGPUIntel(self):
        if self.__getGPUIOKit():
            pass
        else:  # fallback to the default implementation
            gpus = list()
            try:
                gpu_info_dict = json.loads(subprocess.check_output(["system_profiler", "SPDisplaysDataType", "-json"]))
                if 'SPDisplaysDataType' in gpu_info_dict:
                    gpus = gpu_info_dict['SPDisplaysDataType']
                    self.__gpuInfo.number = len(gpus)
                else:
                    pass
            except Exception:
                return

            for gpu in gpus:
                if self.__handleVendor(gpu.get("spdisplays_vendor")) == "Intel":    # Integrated GPU
                    self.__gpuInfo.gpus.append(f'{gpu.get("sppci_model")} [CPU Integrated]')
                elif self.__handleVendor(gpu.get("spdisplays_vendor")) == "AMD":    # dGPU
                    self.__gpuInfo.gpus.append(f'{gpu.get("sppci_model")} {gpu.get("spdisplays_vram")} [Discrete]')
                elif self.__handleVendor(gpu.get("spdisplays_vendor")) == "Nvidia":    # Since current macOS does not support NVIDIA GPUs, this condition is not applicable
                    pass

    def __getGPUIOKit(self):
        try:
            package_root = Path(__file__).resolve().parent.parent.parent
            lib = ctypes.CDLL(f"{package_root}/library/lib/iokitGPULib.dylib")
            lib.getGPUInfo.restype = ctypes.c_char_p
            gpu_info = lib.getGPUInfo()
            gpus = gpu_info.decode('utf-8').split("; ")
            self.__gpuInfo.number = len(gpus)
            for gpu in gpus:
                info_list = gpu.split(", ")
                model = info_list[0]
                vendor_id = info_list[1]
                vram = round(int(info_list[2]) / 1024, None)
                if self.__handleVendorID(vendor_id) == "Intel":    # Integrated GPU
                    self.__gpuInfo.gpus.append(f'{model} [CPU Integrated]')
                elif self.__handleVendorID(vendor_id) == "AMD":    # dGPU
                    self.__gpuInfo.gpus.append(f'{model} {vram} GB [Discrete]')
            return True
        except Exception as e:
            # print(f"An error occurred while getting GPU info using IOKit: {e}")
            return False

    @staticmethod
    def __handleVendor(vendor):
        if vendor == "sppci_vendor_Apple":
            return "Apple"
        elif vendor == "sppci_vendor_intel":
            return "Intel"
        elif vendor == "sppci_vendor_amd":
            return "AMD"
        else:
            return vendor

    @staticmethod
    def __handleVendorID(vendor_id):
        if vendor_id == "0x8086":
            return "Intel"
        elif vendor_id == "0x1002":
            return "AMD"
        else:
            return vendor_id




import re
import os
from .cpuInfo import CPUInfo


class CPUDetectLinux:
    def __init__(self):
        self.__cpuInfo = CPUInfo()

    def getCPUInfo(self):
        self.__getCPUInfo()
        self.__modelClean()
        self.__handleSBC()
        if self.__cpuInfo.model == "":
            self.__cpuInfo.model = "Unknown"
        if self.__cpuInfo.model != "":
            self.__cpuInfo.cpu = self.__cpuInfo.model
            if self.__cpuInfo.cores != "":
                self.__cpuInfo.cpu += f" ({self.__cpuInfo.cores})"
            if self.__cpuInfo.frequency != "":
                self.__cpuInfo.cpu += f" @ {self.__cpuInfo.frequency}"
        return self.__cpuInfo

    def __getCPUInfo(self):
        try:
            with open("/proc/cpuinfo", "r") as f:
                cpu_info = f.read()
        except FileNotFoundError:
            return
        for line in cpu_info.split("\n"):
            if (line.startswith("model name") or line.startswith("Hardware") or line.startswith("Processor") or
                    line.startswith("cpu model") or line.startswith("chip type") or line.startswith("cpu type")):
                model = line.split(":")[1].strip()
                if "@" in model:
                    self.__cpuInfo.model = model.split("@")[0].strip()
                    self.__cpuInfo.frequency = model.split("@")[1].strip()
                else:
                    self.__cpuInfo.model = model
                break
        self.__cpuInfo.cores = len(re.findall(r"processor", cpu_info))
        if self.__cpuInfo.frequency == "":
            if os.path.exists("/sys/devices/system/cpu/cpu0/cpufreq"):
                try:
                    with open("/sys/devices/system/cpu/cpu0/cpufreq/scaling_max_freq", "r") as f:
                        self.__cpuInfo.frequency = f"{int(f.read()) / 1000 / 1000} Ghz"  # Ghz
                except FileNotFoundError:
                    pass
            else:
                for line in cpu_info.split("\n"):
                    if line.startswith("cpu MHz") or line.startswith("clock"):
                        self.__cpuInfo.frequency = f"{round(float(line.split(':')[1].strip()) / 1000, 2)} Ghz"  # Ghz
                        break

    def __modelClean(self):
        self.__cpuInfo.model = self.__cpuInfo.model.replace("(R)", "")
        self.__cpuInfo.model = self.__cpuInfo.model.replace("(TM)", "")

    def __handleSBC(self):
        # some values should be double-checked
        # Info source: https://github.com/raspberrypi/firmware/tree/master/boot
        if os.path.exists("/sys/firmware/devicetree/base/compatible"):
            try:
                with open("/sys/firmware/devicetree/base/compatible", "r") as f:
                    compatible = f.read().strip()
            except FileNotFoundError:
                compatible = ""
            if "raspberrypi" in compatible:
                model = compatible.split(",")[-1]
                if model.startswith("bcm"):
                    self.__cpuInfo.model = model.upper()
            elif "orangepi" in compatible:
                if "allwinner" in compatible:
                    model = compatible.split(",")[-1]
                    if model.startswith("sun"):
                        self.__cpuInfo.model = f"Allwinner {model.split('-')[-1].upper()} ({model})"
            elif "cvitek" in compatible:
                model = compatible.split(",")[-1]
                self.__cpuInfo.model = f"Cvitek {model}"
        else:
            pass





from dataclasses import dataclass
from .colorSet import ColorSet


class ColorConfigSet:
    def __init__(self, os_name):
        self.__os_name = os_name

    def getColorConfigSet(self):
        if self.__os_name == "macOS":
            return ColorConfigSetM.macOS
        elif self.__os_name == "debian":
            return ColorConfigSetD.debian
        elif self.__os_name == "linux":
            return ColorConfigSetL.linux
        elif self.__os_name == "fedora":
            return ColorConfigSetF.fedora
        elif self.__os_name == "fedora_small":
            return ColorConfigSetF.fedora_small
        elif self.__os_name == "ubuntu":
            return ColorConfigSetU.ubuntu
        elif self.__os_name == "ubuntu_small":
            return ColorConfigSetU.ubuntu_small
        elif self.__os_name == "raspbian":
            return ColorConfigSetR.raspbian
        elif self.__os_name == "armbian":
            return ColorConfigSetA.armbian
        elif self.__os_name == "alpine":
            return ColorConfigSetA.alpine
        elif self.__os_name == "arch":
            return ColorConfigSetA.arch
        elif self.__os_name == "centos":
            return ColorConfigSetC.centos
        elif self.__os_name == "freebsd":
            return ColorConfigSetF.freebsd
        elif self.__os_name == "windows_old":
            return ColorConfigSetW.windows_old
        elif self.__os_name == "windows_10":
            return ColorConfigSetW.windows_10
        elif self.__os_name == "windows_11":
            return ColorConfigSetW.windows_11
        elif self.__os_name == "windows_2025":
            return ColorConfigSetW.windows_2025
        else:
            return ColorConfigSetL.linux  # default to Linux


@dataclass
class ColorConfigSetA:
    armbian = {
        "colors": [
            ColorSet.COLOR_FG_RED
        ],
        "colorKeys": ColorSet.COLOR_FG_YELLOW,
        "colorTitle": ColorSet.COLOR_FG_YELLOW
    }
    alpine = {
        "colors": [
            ColorSet.COLOR_FG_BLUE,
            ColorSet.COLOR_FG_WHITE
        ],
        "colorKeys": ColorSet.COLOR_FG_MAGENTA,
        "colorTitle": ColorSet.COLOR_FG_BLUE
    }
    arch = {
        "colors": [
            ColorSet.COLOR_FG_CYAN,
            ColorSet.COLOR_FG_CYAN
        ],
        "colorKeys": ColorSet.COLOR_FG_CYAN,
        "colorTitle": ColorSet.COLOR_FG_CYAN
    }


@dataclass
class ColorConfigSetC:
    centos = {
        "colors": [
            ColorSet.COLOR_FG_YELLOW,
            ColorSet.COLOR_FG_GREEN,
            ColorSet.COLOR_FG_BLUE,
            ColorSet.COLOR_FG_MAGENTA,
            ColorSet.COLOR_FG_WHITE
        ],
        "colorKeys": ColorSet.COLOR_FG_GREEN,
        "colorTitle": ColorSet.COLOR_FG_YELLOW
    }


@dataclass
class ColorConfigSetD:
    debian = {
        "colors": [
            ColorSet.COLOR_FG_RED,
            ColorSet.COLOR_FG_LIGHT_BLACK
        ],
        "colorKeys": ColorSet.COLOR_FG_RED,
        "colorTitle":  ColorSet.COLOR_FG_RED
    }


@dataclass
class ColorConfigSetF:
    fedora = {
        "colors": [
            ColorSet.COLOR_FG_BLUE,
            ColorSet.COLOR_FG_WHITE
        ],
        "colorKeys": ColorSet.COLOR_FG_BLUE,
        "colorTitle":  ColorSet.COLOR_FG_BLUE
    }
    fedora_small = {
        "colors": [
            ColorSet.COLOR_FG_BLUE,
            ColorSet.COLOR_FG_WHITE
        ],
        "colorKeys": ColorSet.COLOR_FG_BLUE,
        "colorTitle": ColorSet.COLOR_FG_BLUE
    }
    freebsd = {
        "colors": [
            ColorSet.COLOR_FG_WHITE,
            ColorSet.COLOR_FG_RED
        ],
        "colorKeys": ColorSet.COLOR_FG_RED,
        "colorTitle": ColorSet.COLOR_FG_RED
    }


@dataclass
class ColorConfigSetL:
    linux = {
        "colors": [
            ColorSet.COLOR_FG_WHITE,
            ColorSet.COLOR_FG_BLACK,
            ColorSet.COLOR_FG_YELLOW
        ],
        "colorKeys": ColorSet.COLOR_FG_YELLOW,
        "colorTitle": ColorSet.COLOR_FG_YELLOW
    }


@dataclass
class ColorConfigSetM:
    macOS = {
        "colors": [
            ColorSet.COLOR_FG_GREEN,
            ColorSet.COLOR_FG_YELLOW,
            ColorSet.COLOR_FG_RED,
            ColorSet.COLOR_FG_MAGENTA,
            ColorSet.COLOR_FG_BLUE
        ],
        "colorKeys": ColorSet.COLOR_FG_YELLOW,
        "colorTitle":  ColorSet.COLOR_FG_GREEN
    }


@dataclass
class ColorConfigSetR:
    raspbian = {
        "colors": [
            ColorSet.COLOR_FG_RED,
            ColorSet.COLOR_FG_GREEN
        ],
        "colorKeys": ColorSet.COLOR_FG_RED,
        "colorTitle": ColorSet.COLOR_FG_GREEN
    }


@dataclass
class ColorConfigSetU:
    ubuntu = {
        "colors": [
            ColorSet.COLOR_FG_RED,
            ColorSet.COLOR_FG_WHITE
        ],
        "colorKeys": ColorSet.COLOR_FG_RED,
        "colorTitle":  ColorSet.COLOR_FG_RED
    }
    ubuntu_small = {
        "colors": [
            ColorSet.COLOR_FG_RED
        ],
        "colorKeys": ColorSet.COLOR_FG_RED,
        "colorTitle": ColorSet.COLOR_FG_RED
    }


@dataclass
class ColorConfigSetW:
    windows_old = {
        "colors": [
            ColorSet.COLOR_FG_RED,
            ColorSet.COLOR_FG_GREEN,
            ColorSet.COLOR_FG_BLUE,
            ColorSet.COLOR_FG_YELLOW
        ],
        "colorKeys": ColorSet.COLOR_FG_BLUE,
        "colorTitle": ColorSet.COLOR_FG_GREEN
    }
    windows_10 = {
        "colors": [
            ColorSet.COLOR_FG_CYAN,
            ColorSet.COLOR_FG_CYAN,
            ColorSet.COLOR_FG_CYAN,
            ColorSet.COLOR_FG_CYAN
        ],
        "colorKeys": ColorSet.COLOR_FG_YELLOW,
        "colorTitle": ColorSet.COLOR_FG_GREEN
    }
    windows_11 = {
        "colors": [
            ColorSet.COLOR_FG_BLUE,
            ColorSet.COLOR_FG_BLUE,
            ColorSet.COLOR_FG_BLUE,
            ColorSet.COLOR_FG_BLUE
        ],
        "colorKeys": ColorSet.COLOR_FG_YELLOW,
        "colorTitle": ColorSet.COLOR_FG_CYAN
    }
    windows_2025 = {
        "colors": [
            ColorSet.COLOR_FG_BLUE,
            ColorSet.COLOR_FG_BLUE,
            ColorSet.COLOR_FG_BLUE,
            ColorSet.COLOR_FG_BLUE
        ],
        "colorKeys": ColorSet.COLOR_FG_YELLOW,
        "colorTitle": ColorSet.COLOR_FG_CYAN
    }



from .frontend import Printer
from .backend import Data
from .backend.title import TitleDetect
from .backend.host import HostDetect
from .backend.kernel import KernelDetect
from .backend.shell import ShellDetect
from .backend.uptime import UptimeDetect
from .backend.os import OSDetect
from .backend.cpu import CPUDetect
from .backend.gpu import GPUDetect
from .backend.memory import MemoryDetect
from .backend.nic import NICDetect
from .backend.npu import NPUDetect
from .pyhwUtil import createDataString
from .pyhwUtil import getOS, selectOSLogo
import multiprocessing


def detect_title(os, result_dict):
    result_dict["title"] = TitleDetect(os=os).getTitle().title


def detect_host(os, result_dict):
    result_dict["Host"] = HostDetect(os=os).getHostInfo().model


def detect_kernel(os, result_dict):
    result_dict["Kernel"] = KernelDetect(os=os).getKernelInfo().kernel


def detect_shell(os, result_dict):
    result_dict["Shell"] = ShellDetect(os=os).getShellInfo().info


def detect_uptime(os, result_dict):
    result_dict["Uptime"] = UptimeDetect(os=os).getUptime().uptime


def detect_os(os, result_dict):
    result_dict["OS"] = OSDetect(os=os).getOSInfo().prettyName


def detect_cpu(os, result_dict):
    result_dict["CPU"] = CPUDetect(os=os).getCPUInfo().cpu


def detect_gpu(os, result_dict):
    gpu_info = GPUDetect(os=os).getGPUInfo()
    if gpu_info.number > 0:
        result_dict["GPU"] = gpu_info.gpus


def detect_memory(os, result_dict):
    result_dict["Memory"] = MemoryDetect(os=os).getMemoryInfo().memory


def detect_nic(os, result_dict):
    nic_info = NICDetect(os=os).getNICInfo()
    if nic_info.number > 0:
        result_dict["NIC"] = nic_info.nics


def detect_npu(os, result_dict):
    npu_info = NPUDetect(os=os).getNPUInfo()
    if npu_info.number > 0:
        result_dict["NPU"] = npu_info.npus


def main():
    current_os = getOS()
    if current_os not in ["linux", "macos", "freebsd", "windows"]:
        print(f"Only Linux, macOS, FreeBSD, and Windows are supported for now. Current OS: {current_os}")
        return

    data = Data()

    manager = multiprocessing.Manager()
    result_dict = manager.dict()

    processes = [
        multiprocessing.Process(target=detect_title, args=(current_os, result_dict)),
        multiprocessing.Process(target=detect_host, args=(current_os, result_dict)),
        multiprocessing.Process(target=detect_kernel, args=(current_os, result_dict)),
        multiprocessing.Process(target=detect_shell, args=(current_os, result_dict)),
        multiprocessing.Process(target=detect_uptime, args=(current_os, result_dict)),
        multiprocessing.Process(target=detect_os, args=(current_os, result_dict)),
        multiprocessing.Process(target=detect_cpu, args=(current_os, result_dict)),
        multiprocessing.Process(target=detect_gpu, args=(current_os, result_dict)),
        multiprocessing.Process(target=detect_memory, args=(current_os, result_dict)),
        multiprocessing.Process(target=detect_nic, args=(current_os, result_dict)),
        multiprocessing.Process(target=detect_npu, args=(current_os, result_dict)),
    ]

    for process in processes:
        process.start()

    for process in processes:
        process.join()

    for key, value in result_dict.items():
        setattr(data, key, value)

    logo_os = selectOSLogo(OSDetect(os=current_os).getOSInfo().id)
    Printer(logo_os=logo_os, data=createDataString(data)).cPrint()


# def main():
#     current_os = getOS()
#     if current_os not in ["linux", "macos", "freebsd", "windows"]:
#         print(f"Only Linux, macOS, FreeBSD and Windows are supported for now. Current OS: {current_os}")
#         return
#     data = Data()
#     data.title = TitleDetect(os=current_os).getTitle().title
#     data.Host = HostDetect(os=current_os).getHostInfo().model
#     data.Kernel = KernelDetect(os=current_os).getKernelInfo().kernel
#     data.Shell = ShellDetect(os=current_os).getShellInfo().info
#     data.Uptime = UptimeDetect(os=current_os).getUptime().uptime
#     data.OS = OSDetect(os=current_os).getOSInfo().prettyName
#     data.CPU = CPUDetect(os=current_os).getCPUInfo().cpu
#     gpu_info = GPUDetect(os=current_os).getGPUInfo()
#     if gpu_info.number > 0:
#         data.GPU = gpu_info.gpus
#     data.Memory = MemoryDetect(os=current_os).getMemoryInfo().memory
#     nic_info = NICDetect(os=current_os).getNICInfo()
#     if nic_info.number > 0:
#         data.NIC = nic_info.nics
#     npu_info = NPUDetect(os=current_os).getNPUInfo()
#     if npu_info.number > 0:
#         data.NPU = npu_info.npus
#
#     Printer(logo_os=selectOSLogo(OSDetect(os=current_os).getOSInfo().id), data=createDataString(data)).cPrint()


if __name__ == "__main__":
    main()



"""
    In dev.
"""
from .hostInfo import HostInfo
from ...pyhwUtil import sysctlGetString


class HostDetectMacOS:
    def __init__(self):
        self.__hostInfo = HostInfo()
        self.__HWModel = ""

    def getHostInfo(self):
        self.__getHWModel()
        self.__hostInfo.model = self.__handleMacName(self.__HWModel)
        return self.__hostInfo

    def __getHWModel(self):
        self.__HWModel = sysctlGetString("hw.model")

    def __handleMacName(self, hw_model: str):
        # This part of the code is directly re-implemented from the fastfetch project. It seems that there is no
        # more elegant way to get the product name of the Mac.
        # See https://github.com/fastfetch-cli/fastfetch/blob/dev/src/detection/host/host_apple.c for more details.
        if hw_model.startswith("MacBookPro"):
            version = hw_model[len("MacBookPro"):]
            if self.__hwModelCheck(version, "18,3") or self.__hwModelCheck(version, "18,4"):
                return "MacBook Pro (14-inch, 2021)"
            if self.__hwModelCheck(version, "18,1") or self.__hwModelCheck(version, "18,2"):
                return "MacBook Pro (16-inch, 2021)"
            if self.__hwModelCheck(version, "17,1"):
                return "MacBook Pro (13-inch, M1, 2020)"
            if self.__hwModelCheck(version, "16,3"):
                return "MacBook Pro (13-inch, 2020, Two Thunderbolt 3 ports)"
            if self.__hwModelCheck(version, "16,2"):
                return "MacBook Pro (13-inch, 2020, Four Thunderbolt 3 ports)"
            if self.__hwModelCheck(version, "16,4") or self.__hwModelCheck(version, "16,1"):
                return "MacBook Pro (16-inch, 2019)"
            if self.__hwModelCheck(version, "15,4"):
                return "MacBook Pro (13-inch, 2019, Two Thunderbolt 3 ports)"
            if self.__hwModelCheck(version, "15,3"):
                return "MacBook Pro (15-inch, 2019)"
            if self.__hwModelCheck(version, "15,2"):
                return "MacBook Pro (13-inch, 2018/2019, Four Thunderbolt 3 ports)"
            if self.__hwModelCheck(version, "15,1"):
                return "MacBook Pro (15-inch, 2018/2019)"
            if self.__hwModelCheck(version, "14,3"):
                return "MacBook Pro (15-inch, 2017)"
            if self.__hwModelCheck(version, "14,2"):
                return "MacBook Pro (13-inch, 2017, Four Thunderbolt 3 ports)"
            if self.__hwModelCheck(version, "14,1"):
                return "MacBook Pro (13-inch, 2017, Two Thunderbolt 3 ports)"
            if self.__hwModelCheck(version, "13,3"):
                return "MacBook Pro (15-inch, 2016)"
            if self.__hwModelCheck(version, "13,2"):
                return "MacBook Pro (13-inch, 2016, Four Thunderbolt 3 ports)"
            if self.__hwModelCheck(version, "13,1"):
                return "MacBook Pro (13-inch, 2016, Two Thunderbolt 3 ports)"
            if self.__hwModelCheck(version, "12,1"):
                return "MacBook Pro (Retina, 13-inch, Early 2015)"
            if self.__hwModelCheck(version, "11,4") or self.__hwModelCheck(version, "11,5"):
                return "MacBook Pro (Retina, 15-inch, Mid 2015)"
            if self.__hwModelCheck(version, "11,2") or self.__hwModelCheck(version, "11,3"):
                return "MacBook Pro (Retina, 15-inch, Late 2013/Mid 2014)"
            if self.__hwModelCheck(version, "11,1"):
                return "MacBook Pro (Retina, 13-inch, Late 2013/Mid 2014)"
            if self.__hwModelCheck(version, "10,2"):
                return "MacBook Pro (Retina, 13-inch, Late 2012/Early 2013)"
            if self.__hwModelCheck(version, "10,1"):
                return "MacBook Pro (Retina, 15-inch, Mid 2012/Early 2013)"
            if self.__hwModelCheck(version, "9,2"):
                return "MacBook Pro (13-inch, Mid 2012)"
            if self.__hwModelCheck(version, "9,1"):
                return "MacBook Pro (15-inch, Mid 2012)"
            if self.__hwModelCheck(version, "8,3"):
                return "MacBook Pro (17-inch, 2011)"
            if self.__hwModelCheck(version, "8,2"):
                return "MacBook Pro (15-inch, 2011)"
            if self.__hwModelCheck(version, "8,1"):
                return "MacBook Pro (13-inch, 2011)"
            if self.__hwModelCheck(version, "7,1"):
                return "MacBook Pro (13-inch, Mid 2010)"
            if self.__hwModelCheck(version, "6,2"):
                return "MacBook Pro (15-inch, Mid 2010)"
            if self.__hwModelCheck(version, "6,1"):
                return "MacBook Pro (17-inch, Mid 2010)"
            if self.__hwModelCheck(version, "5,5"):
                return "MacBook Pro (13-inch, Mid 2009)"
            if self.__hwModelCheck(version, "5,3"):
                return "MacBook Pro (15-inch, Mid 2009)"
            if self.__hwModelCheck(version, "5,2"):
                return "MacBook Pro (17-inch, Mid/Early 2009)"
            if self.__hwModelCheck(version, "5,1"):
                return "MacBook Pro (15-inch, Late 2008)"
            if self.__hwModelCheck(version, "4,1"):
                return "MacBook Pro (17/15-inch, Early 2008)"
        elif hw_model.startswith("MacBookAir"):
            version = hw_model[len("MacBookAir"):]
            if self.__hwModelCheck(version, "10,1"):
                return "MacBook Air (M1, 2020)"
            if self.__hwModelCheck(version, "9,1"):
                return "MacBook Air (Retina, 13-inch, 2020)"
            if self.__hwModelCheck(version, "8,2"):
                return "MacBook Air (Retina, 13-inch, 2019)"
            if self.__hwModelCheck(version, "8,1"):
                return "MacBook Air (Retina, 13-inch, 2018)"
            if self.__hwModelCheck(version, "7,2"):
                return "MacBook Air (13-inch, Early 2015/2017)"
            if self.__hwModelCheck(version, "7,1"):
                return "MacBook Air (11-inch, Early 2015)"
            if self.__hwModelCheck(version, "6,2"):
                return "MacBook Air (13-inch, Mid 2013/Early 2014)"
            if self.__hwModelCheck(version, "6,1"):
                return "MacBook Air (11-inch, Mid 2013/Early 2014)"
            if self.__hwModelCheck(version, "5,2"):
                return "MacBook Air (13-inch, Mid 2012)"
            if self.__hwModelCheck(version, "5,1"):
                return "MacBook Air (11-inch, Mid 2012)"
            if self.__hwModelCheck(version, "4,2"):
                return "MacBook Air (13-inch, Mid 2011)"
            if self.__hwModelCheck(version, "4,1"):
                return "MacBook Air (11-inch, Mid 2011)"
            if self.__hwModelCheck(version, "3,2"):
                return "MacBook Air (13-inch, Late 2010)"
            if self.__hwModelCheck(version, "3,1"):
                return "MacBook Air (11-inch, Late 2010)"
            if self.__hwModelCheck(version, "2,1"):
                return "MacBook Air (Mid 2009)"

        elif hw_model.startswith("Macmini"):
            version = hw_model[len("Macmini"):]
            if self.__hwModelCheck(version, "9,1"):
                return "Mac mini (M1, 2020)"
            if self.__hwModelCheck(version, "8,1"):
                return "Mac mini (2018)"
            if self.__hwModelCheck(version, "7,1"):
                return "Mac mini (Mid 2014)"
            if self.__hwModelCheck(version, "6,1") or self.__hwModelCheck(version, "6,2"):
                return "Mac mini (Late 2012)"
            if self.__hwModelCheck(version, "5,1") or self.__hwModelCheck(version, "5,2"):
                return "Mac mini (Mid 2011)"
            if self.__hwModelCheck(version, "4,1"):
                return "Mac mini (Mid 2010)"
            if self.__hwModelCheck(version, "3,1"):
                return "Mac mini (Early/Late 2009)"

        elif hw_model.startswith("MacBook"):
            version = hw_model[len("MacBook"):]
            if self.__hwModelCheck(version, "10,1"):
                return "MacBook (Retina, 12-inch, 2017)"
            if self.__hwModelCheck(version, "9,1"):
                return "MacBook (Retina, 12-inch, Early 2016)"
            if self.__hwModelCheck(version, "8,1"):
                return "MacBook (Retina, 12-inch, Early 2015)"
            if self.__hwModelCheck(version, "7,1"):
                return "MacBook (13-inch, Mid 2010)"
            if self.__hwModelCheck(version, "6,1"):
                return "MacBook (13-inch, Late 2009)"
            if self.__hwModelCheck(version, "5,2"):
                return "MacBook (13-inch, Early/Mid 2009)"

        elif hw_model.startswith("MacPro"):
            version = hw_model[len("MacPro"):]
            if self.__hwModelCheck(version, "7,1"):
                return "Mac Pro (2019)"
            if self.__hwModelCheck(version, "6,1"):
                return "Mac Pro (Late 2013)"
            if self.__hwModelCheck(version, "5,1"):
                return "Mac Pro (Mid 2010 - Mid 2012)"
            if self.__hwModelCheck(version, "4,1"):
                return "Mac Pro (Early 2009)"

        elif hw_model.startswith("Mac"):
            version = hw_model[len("Mac"):]
            if self.__hwModelCheck(version, "16,3"):
                return "iMac (24-inch, 2024, Four Thunderbolt / USB 4 ports)"
            if self.__hwModelCheck(version, "16,2"):
                return "iMac (24-inch, 2024, Two Thunderbolt / USB 4 ports)"
            if self.__hwModelCheck(version, "16,1") or self.__hwModelCheck(version, "16,6") or self.__hwModelCheck(version, "16,8"):
                return "MacBook Pro (14-inch, 2024, Three Thunderbolt 4 ports)"
            if self.__hwModelCheck(version, "16,5") or self.__hwModelCheck(version, "16,7"):
                return "MacBook Pro (16-inch, 2024, Three Thunderbolt 4 ports)"
            if self.__hwModelCheck(version, "16,10") or self.__hwModelCheck(version, "16,15"):
                return "Mac mini (M4, 2024)"
            if self.__hwModelCheck(version, "15,13"):
                return "MacBook Air (15-inch, M3, 2024)"
            if self.__hwModelCheck(version, "15,2"):
                return "MacBook Air (13-inch, M3, 2024)"
            if self.__hwModelCheck(version, "15,3"):
                return "MacBook Pro (14-inch, Nov 2023, Two Thunderbolt / USB 4 ports)"
            if self.__hwModelCheck(version, "15,4"):
                return "iMac (24-inch, 2023, Two Thunderbolt / USB 4 ports)"
            if self.__hwModelCheck(version, "15,5"):
                return "iMac (24-inch, 2023, Two Thunderbolt / USB 4 ports, Two USB 3 ports)"
            if self.__hwModelCheck(version, "15,6") or self.__hwModelCheck(version, "15,8") or self.__hwModelCheck(version, "15,10"):
                return "MacBook Pro (14-inch, Nov 2023, Three Thunderbolt 4 ports)"
            if self.__hwModelCheck(version, "15,7") or self.__hwModelCheck(version, "15,9") or self.__hwModelCheck(version, "15,11"):
                return "MacBook Pro (16-inch, Nov 2023, Three Thunderbolt 4 ports)"
            if self.__hwModelCheck(version, "14,15"):
                return "MacBook Air (15-inch, M2, 2023)"
            if self.__hwModelCheck(version, "14,14"):
                return "Mac Studio (M2 Ultra, 2023, Two Thunderbolt 4 front ports)"
            if self.__hwModelCheck(version, "14,13"):
                return "Mac Studio (M2 Max, 2023, Two USB-C front ports)"
            if self.__hwModelCheck(version, "14,8"):
                return "Mac Pro (2023)"
            if self.__hwModelCheck(version, "14,6") or self.__hwModelCheck(version, "14,10"):
                return "MacBook Pro (16-inch, 2023)"
            if self.__hwModelCheck(version, "14,5") or self.__hwModelCheck(version, "14,9"):
                return "MacBook Pro (14-inch, 2023)"
            if self.__hwModelCheck(version, "14,3"):
                return "Mac mini (M2, 2023, Two Thunderbolt 4 ports)"
            if self.__hwModelCheck(version, "14,12"):
                return "Mac mini (M2, 2023, Four Thunderbolt 4 ports)"
            if self.__hwModelCheck(version, "14,7"):
                return "MacBook Pro (13-inch, M2, 2022)"
            if self.__hwModelCheck(version, "14,2"):
                return "MacBook Air (M2, 2022)"
            if self.__hwModelCheck(version, "13,1"):
                return "Mac Studio (M1 Max, 2022, Two USB-C front ports)"
            if self.__hwModelCheck(version, "13,2"):
                return "Mac Studio (M1 Ultra, 2022, Two Thunderbolt 4 front ports)"

        elif hw_model.startswith("iMac"):
            version = hw_model[len("iMac"):]
            if self.__hwModelCheck(version, "21,1"):
                return "iMac (24-inch, M1, 2021, Two Thunderbolt / USB 4 ports, Two USB 3 ports)"
            if self.__hwModelCheck(version, "21,2"):
                return "iMac (24-inch, M1, 2021, Two Thunderbolt / USB 4 ports)"
            if self.__hwModelCheck(version, "20,1") or self.__hwModelCheck(version, "20,2"):
                return "iMac (Retina 5K, 27-inch, 2020)"
            if self.__hwModelCheck(version, "19,1"):
                return "iMac (Retina 5K, 27-inch, 2019)"
            if self.__hwModelCheck(version, "19,2"):
                return "iMac (Retina 4K, 21.5-inch, 2019)"
            if self.__hwModelCheck(version, "Pro1,1"):
                return "iMac Pro (2017)"
            if self.__hwModelCheck(version, "18,3"):
                return "iMac (Retina 5K, 27-inch, 2017)"
            if self.__hwModelCheck(version, "18,2"):
                return "iMac (Retina 4K, 21.5-inch, 2017)"
            if self.__hwModelCheck(version, "18,1"):
                return "iMac (21.5-inch, 2017)"
            if self.__hwModelCheck(version, "17,1"):
                return "iMac (Retina 5K, 27-inch, Late 2015)"
            if self.__hwModelCheck(version, "16,2"):
                return "iMac (Retina 4K, 21.5-inch, Late 2015)"
            if self.__hwModelCheck(version, "16,1"):
                return "iMac (21.5-inch, Late 2015)"
            if self.__hwModelCheck(version, "15,1"):
                return "iMac (Retina 5K, 27-inch, Late 2014 - Mid 2015)"
            if self.__hwModelCheck(version, "14,4"):
                return "iMac (21.5-inch, Mid 2014)"
            if self.__hwModelCheck(version, "14,2"):
                return "iMac (27-inch, Late 2013)"
            if self.__hwModelCheck(version, "14,1"):
                return "iMac (21.5-inch, Late 2013)"
            if self.__hwModelCheck(version, "13,2"):
                return "iMac (27-inch, Late 2012)"
            if self.__hwModelCheck(version, "13,1"):
                return "iMac (21.5-inch, Late 2012)"
            if self.__hwModelCheck(version, "12,2"):
                return "iMac (27-inch, Mid 2011)"
            if self.__hwModelCheck(version, "12,1"):
                return "iMac (21.5-inch, Mid 2011)"
            if self.__hwModelCheck(version, "11,3"):
                return "iMac (27-inch, Mid 2010)"
            if self.__hwModelCheck(version, "11,2"):
                return "iMac (21.5-inch, Mid 2010)"
            if self.__hwModelCheck(version, "10,1"):
                return "iMac (27/21.5-inch, Late 2009)"
            if self.__hwModelCheck(version, "9,1"):
                return "iMac (24/20-inch, Early 2009)"

        return hw_model

    @staticmethod
    def __hwModelCheck(version: str, target: str):
        return version == target



from .npuInfo import NPUInfo
from ...pyhwUtil import getArch, getOS
import json
import subprocess
from ..cpu import CPUDetect


class NPUDetectMacOS:
    def __init__(self):
        self.__npuInfo = NPUInfo()
        self.__arch = getArch()

    def getNPUInfo(self):
        self.__npuInfo.npus.append(self.__getNPUbyModelName())
        self.__npuInfo.number += 1
        # if self.__arch == "aarch64":
        #     self.__getNPUAppleSilicon()
        # else:   # Does not consider powerPC based Macs.
        #     self.__getNPUIntel()
        return self.__npuInfo

    def __getNPUAppleSilicon(self):
        # Place holder
        self.__npuInfo.npus.append("Apple Neural Engine [SOC Integrated]")
        self.__npuInfo.number += 1

    def __getNPUIntel(self):
        # Place holder
        self.__npuInfo.npus.append("Not Found")
        self.__npuInfo.number += 1

    @staticmethod
    def __handleVendor(vendor):
        if vendor == "sppci_vendor_Apple":
            return "Apple"
        elif vendor == "sppci_vendor_intel":
            return "Intel"
        elif vendor == "sppci_vendor_amd":
            return "AMD"
        else:
            return vendor

    @staticmethod
    def __getNPUbyModelName():
        # Placeholder
        # see https://apple.fandom.com/wiki/Neural_Engine for more details.
        model_name = CPUDetect(os=getOS()).getCPUInfo().model
        npu = {
            "Apple M1": "Apple Neural Engine 16 Core [SOC Integrated]",
            "Apple M1 Pro": "Apple Neural Engine 16 Core [SOC Integrated]",
            "Apple M1 Max": "Apple Neural Engine 16 Core [SOC Integrated]",
            "Apple M1 Ultra": "Apple Neural Engine 32 Core [SOC Integrated]",
            "Apple M2": "Apple Neural Engine 16 Core [SOC Integrated]",
            "Apple M2 Pro": "Apple Neural Engine 16 Core [SOC Integrated]",
            "Apple M2 Max": "Apple Neural Engine 16 Core [SOC Integrated]",
            "Apple M2 Ultra": "Apple Neural Engine 32 Core [SOC Integrated]",
            "Apple M3": "Apple Neural Engine 16 Core [SOC Integrated]",
            "Apple M3 Pro": "Apple Neural Engine 16 Core [SOC Integrated]",
            "Apple M3 Max": "Apple Neural Engine 16 Core [SOC Integrated]",
            "Apple M4": "Apple Neural Engine 16 Core [SOC Integrated]",
            "Apple M4 Pro": "Apple Neural Engine 16 Core [SOC Integrated]",
            "Apple M4 Max": "Apple Neural Engine 16 Core [SOC Integrated]"
        }
        return npu.get(model_name, "Not Found")




from dataclasses import dataclass


@dataclass
class OSInfo:
    """
    Dataclass to hold the OS information
    """
    def __init__(self):
        """
        Initialize the dataclass, prettyName and id attributes are required.
        """
        self.prettyName = ""
        self.name = ""
        self.id = ""
        self.idLike = ""
        self.variant = ""
        self.variantID = ""
        self.version = ""
        self.versionID = ""
        self.versionCodename = ""
        self.codeName = ""
        self.buildID = ""



from .logo import Logo
from .color import ColorConfigSet, colorPrefix, colorSuffix, ColorSet
from ..pyhwUtil import getOS
from ..pyhwException import BackendException
import os
import re


class Printer:
    def __init__(self, logo_os: str, data: str):
        self.__logo = Logo(logo_os).getLogoContent()
        self.__data = data
        self.__config = ColorConfigSet(logo_os).getColorConfigSet()
        self.__logo_lines = self.__logo.split("\n")
        self.__data_lines = self.__data.strip().split("\n")
        self.__line_length = []
        self.__processed_logo_lines = []
        self.__processed_data_lines = []
        self.__combined_lines = []
        self.__logo_color_indexes = {}
        self.__reg = r'\$(\d)'
        self.__columns = self.__getColumns()

    def cPrint(self):
        self.__LogoPreprocess()
        self.__DataPreprocess()
        max_len_logo = max(self.__line_length)
        for i, (logo_line, data_line) in enumerate(zip(self.__processed_logo_lines, self.__processed_data_lines)):
            combined_line = logo_line + " " * (max_len_logo - self.__line_length[i] + 1) + data_line
            self.__combined_lines.append(combined_line)

        for i, logo_line in enumerate(self.__processed_logo_lines[len(self.__processed_data_lines):], start=len(self.__processed_data_lines)):
            self.__combined_lines.append(logo_line)

        for data_line in self.__processed_data_lines[len(self.__processed_logo_lines):]:
            self.__combined_lines.append(" " * (max_len_logo + 1) + data_line)

        self.__dropLongString()

        print("\n".join(self.__combined_lines))

    def __dropLongString(self):
        # Need more accurate way to drop long strings
        if getOS() == "linux":
            fixed_lines = list()
            for line in self.__combined_lines:
                if len(line) > self.__columns+20:
                    fixed_lines.append(line[:self.__columns+20])
                else:
                    fixed_lines.append(line)
            self.__combined_lines = fixed_lines
        else:
            pass


    @staticmethod
    def __getColumns() -> int:
        if getOS() == "linux":
            try:
                _, columns_str = os.popen('stty size', 'r').read().split()
                columns = int(columns_str)
            except:
                columns = 80  # default terminal size is 80 columns
        else:
            # macOS default terminal size is 80 columns
            columns = 80
        return columns

    def __LogoPreprocess(self):
        global_color = self.__config.get("colors")[0]
        for logo_line in self.__logo_lines:
            matches = re.findall(pattern=self.__reg, string=logo_line)
            color_numbers = len(matches)
            line_length = len(logo_line)
            if color_numbers > 0:
                colors = [int(match) - 1 for match in matches]  # color indexes
                temp_line = colorPrefix(ColorSet.COLOR_MODE_BOLD) + colorPrefix(global_color) + logo_line + colorSuffix()
                for color in colors:
                    temp_line = temp_line.replace(f"${color+1}", colorPrefix(self.__config.get("colors")[color]))
                global_color = self.__config.get("colors")[colors[-1]]  # set the global color to the last used color
            else:
                temp_line = colorPrefix(ColorSet.COLOR_MODE_BOLD) + colorPrefix(global_color) + logo_line + colorSuffix()
            flags_number = len(re.findall(pattern=r'\$\$', string=logo_line)) + len(re.findall(pattern=r'\$\0', string=logo_line))
            if flags_number > 0:
                temp_line = temp_line.replace("$$", "$")
                temp_line = temp_line.replace("$\0", "$")
            self.__line_length.append(line_length - 2 * color_numbers - flags_number)
            self.__processed_logo_lines.append(temp_line)

    def __DataPreprocess(self):
        header_color = self.__config.get("colorTitle")
        keys_color = self.__config.get("colorKeys")
        self.__processed_data_lines.append(" " + colorPrefix(ColorSet.COLOR_MODE_BOLD) + colorPrefix(header_color) +
                                           self.__data_lines[0].split("@")[0] + colorSuffix() + colorPrefix(ColorSet.COLOR_MODE_BOLD) +
                                           "@" + colorPrefix(header_color) +
                                           self.__data_lines[0].split("@")[1] + colorSuffix())
        self.__processed_data_lines.append(colorSuffix() + self.__data_lines[1])
        for data_line in self.__data_lines[2:]:
            try:
                spl = data_line.split(": ")
                name = spl[0]
                value = "".join(spl[1:])
            except:
                print(data_line)
                raise BackendException("Invalid data format")
            self.__processed_data_lines.append(colorPrefix(ColorSet.COLOR_MODE_BOLD) + colorPrefix(keys_color) + name + ": " + colorSuffix() + value)





import ctypes


mylib = ctypes.CDLL('./metalGPULib.dylib')

mylib.backend_init.argtypes = []
mylib.backend_init.restype = None

mylib.get_default_device_name.argtypes = []
mylib.get_default_device_name.restype = ctypes.c_char_p


# Call the function
mylib.backend_init()
device_name = mylib.get_default_device_name()

# Print the result
print(f"Device Name: {device_name.decode('utf-8')}")




"""
    In dev.
"""
from ...pyhwException import BackendException
from .shellInfo import ShellInfo
import json
import subprocess


class ShellDetectWindows:
    def __init__(self):
        self.__shellInfo = ShellInfo()

    def getShellInfo(self):
        COMMAND = "$PSVersionTable.PSVersion | ConvertTo-JSON"

        try:
            result = subprocess.run(["powershell", "-NoProfile", "-Command", COMMAND], capture_output=True, text=True)
        except subprocess.SubprocessError:
            raise BackendException("Error running PowerShell command.")

        res = json.loads(result.stdout)
        major = res["Major"]
        minor = res["Minor"]

        self.__shellInfo.shell = "PowerShell"
        self.__shellInfo.version = f"{major}.{minor}"
        self.__shellInfo.info = f"PowerShell {major}.{minor}"
        return self.__shellInfo



from .cpuInfo import CPUInfo
from ...pyhwUtil import sysctlGetString, sysctlGetInt, getArch


class CPUDetectBSD:
    def __init__(self):
        self.__cpuInfo = CPUInfo()

    def getCPUInfo(self):
        self.__cpuInfo.cpu = sysctlGetString("hw.model")
        return self.__cpuInfo




from ...pyhwException import OSUnsupportedException


class ShellDetect:
    def __init__(self, os):
        self.OS = os

    def getShellInfo(self):
        if self.OS in ["linux", "macos", "freebsd"]:
            from .unix import ShellDetectUnix
            return ShellDetectUnix().getShellInfo()
        elif self.OS == "windows":
            from .windows import ShellDetectWindows
            return ShellDetectWindows().getShellInfo()
        else:
            raise OSUnsupportedException("Unsupported operating system")



from ...pyhwException import OSUnsupportedException


class NPUDetect:
    def __init__(self, os):
        self.OS = os

    def getNPUInfo(self):
        if self.OS == "linux":
            from .linux import NPUDetectLinux
            return NPUDetectLinux().getNPUInfo()
        elif self.OS == "macos":
            from .macos import NPUDetectMacOS
            return NPUDetectMacOS().getNPUInfo()
        elif self.OS == "freebsd":
            from .bsd import NPUDetectBSD
            return NPUDetectBSD().getNPUInfo()
        elif self.OS == "windows":
            from .windows import NPUDetectWindows
            return NPUDetectWindows().getNPUInfo()
        else:
            raise OSUnsupportedException("Unsupported operating system")



from ...pyhwException import OSUnsupportedException


class HostDetect:
    def __init__(self, os):
        self.OS = os

    def getHostInfo(self):
        if self.OS == "linux":
            from .linux import HostDetectLinux
            return HostDetectLinux().getHostInfo()
        elif self.OS == "macos":
            from .macos import HostDetectMacOS
            return HostDetectMacOS().getHostInfo()
        elif self.OS == "freebsd":
            from .bsd import HostDetectBSD
            return HostDetectBSD().getHostInfo()
        elif self.OS == "windows":
            from .windows import HostDetectWindows
            return HostDetectWindows().getHostInfo()
        else:
            raise OSUnsupportedException("Unsupported operating system")



"""
    In dev.
"""
from .osInfo import OSInfo


class OSDetectLinux:
    def __init__(self):
        self._osInfo = OSInfo()

    def getOSInfo(self):
        """
        Detects the os distribution and its version.
        :return: dataclass OSInfoLinux, direct attrs: prettyName
        """
        self._getOSInfo()
        self._handleArmbian()
        return self._osInfo

    def _getOSInfo(self):
        try:
            with open("/etc/os-release", "r") as f:
                for line in f:
                    key, value = line.strip().split("=")
                    if key == "PRETTY_NAME":
                        self._osInfo.prettyName = value.strip('"')
                    elif key == "NAME":
                        self._osInfo.name = value.strip('"')
                    elif key == "ID":
                        self._osInfo.id = value.strip('"')
                    elif key == "ID_LIKE":
                        self._osInfo.idLike = value.strip('"')
                    elif key == "VARIANT":
                        self._osInfo.variant = value.strip('"')
                    elif key == "VARIANT_ID":
                        self._osInfo.variantID = value.strip('"')
                    elif key == "VERSION":
                        self._osInfo.version = value.strip('"')
                    elif key == "VERSION_ID":
                        self._osInfo.versionID = value.strip('"')
                    elif key == "VERSION_CODENAME":
                        self._osInfo.versionCodename = value.strip('"')
                    elif key == "CODE_NAME":
                        self._osInfo.codeName = value.strip('"')
                    elif key == "BUILD_ID":
                        self._osInfo.buildID = value.strip('"')
        except Exception:
            pass

    def _handleArmbian(self):
        if "Armbian" in self._osInfo.prettyName or "armbian" in self._osInfo.prettyName:
            self._osInfo.id = "armbian"



from ...pyhwException import BackendException
from .uptimeInfo import UptimeInfo
import subprocess
import json
import re
from datetime import datetime


class UptimeDetectWindows:
    def __init__(self):
        self.__uptimeInfo = UptimeInfo()

    def getUptimeInfo(self):
        self.__getUptime()
        return self.__uptimeInfo

    def __getUptime(self):
        COMMAND = 'Get-CimInstance -ClassName Win32_OperatingSystem | Select-Object LastBootUpTime | ConvertTo-JSON'
        try:
            result = subprocess.run(["powershell", "-NoProfile", "-Command", COMMAND], capture_output=True, text=True)
        except subprocess.SubprocessError:
            raise BackendException("Error while getting system uptime")

        res = json.loads(result.stdout)
        timestamp_str = res['LastBootUpTime']
        match = re.search(r'\d+', timestamp_str)

        if match:
            timestamp_ms = int(match.group(0))
            last_boot_time = datetime.utcfromtimestamp(timestamp_ms / 1000)
            current_time = datetime.utcnow()
            uptime = current_time - last_boot_time
            days = uptime.days
            seconds = uptime.seconds
            hours = seconds // 3600
            minutes = (seconds % 3600) // 60
            secs = (seconds % 3600) % 60

            if days == 0:
                if hours == 0:
                    self.__uptimeInfo.uptime = f"{minutes} mins {secs} secs"
                else:
                    self.__uptimeInfo.uptime = f"{hours} hours {minutes} mins {secs} secs"
            else:
                self.__uptimeInfo.uptime = f"{days} days {hours} hours {minutes} mins {secs} secs"
        else:
            raise BackendException("Error while getting system uptime")





import subprocess
from .gpuInfo import GPUInfo
from ..cpu import CPUDetect
from ...pyhwUtil import getArch
import pypci


class GPUDetectLinux:
    def __init__(self):
        self.__gpuInfo = GPUInfo()

    def getGPUInfo(self):
        self.__getGPUInfo()
        self.__sortGPUList()
        return self.__gpuInfo

    def __getGPUInfo(self):
        gpu_devices = pypci.PCI().FindAllVGA()
        if len(gpu_devices) == 0:
            self.__handleNonePciDevices()
        else:
            for device in gpu_devices:
                if device.subsystem_device_name != "":
                    device_name = f"{device.vendor_name} {device.device_name} ({device.subsystem_device_name})"
                else:
                    device_name = f"{device.vendor_name} {device.device_name}"
                self.__gpuInfo.gpus.append(self.__gpuNameClean(device_name))
                self.__gpuInfo.number += 1

    def __handleNonePciDevices(self):
        # if detector can't find any VGA/Display/3D GPUs, assume the host is a sbc device, this function is a placeholder for a more advanced method.
        if getArch() in ["aarch64", "arm32", "riscv64"]:
            self.__gpuInfo.number = 1
            self.__gpuInfo.gpus.append(f"{CPUDetect(os='linux').getCPUInfo().model} [SOC Integrated]")
        else:
            self.__gpuInfo.number = 1
            self.__gpuInfo.gpus.append("Not found")

    @staticmethod
    def __gpuNameClean(gpu_name: str):
        gpu_name_clean = gpu_name.replace("Corporation ", "")
        return gpu_name_clean

    def __sortGPUList(self):
        self.__gpuInfo.gpus.sort()




from .kernelInfo import KernelInfo
import subprocess


class KernelDetectUnix:
    def __init__(self):
        self.__kernelInfo = KernelInfo()

    def getKernelInfo(self):
        self.__getKernelName()
        self.__getKernelVersion()
        self.__getKernelMachine()
        self.__kernelInfo.kernel = self.__kernelInfo.name + " " + self.__kernelInfo.version + " " + self.__kernelInfo.machine
        return self.__kernelInfo

    def __getKernelName(self):
        try:
            result = subprocess.run(['uname', '-s'], capture_output=True, text=True)
            self.__kernelInfo.name = result.stdout.strip()
        except subprocess.SubprocessError:
            pass

    def __getKernelVersion(self):
        try:
            result = subprocess.run(['uname', '-r'], capture_output=True, text=True)
            self.__kernelInfo.version = result.stdout.strip()
        except subprocess.SubprocessError:
            pass

    def __getKernelMachine(self):
        try:
            result = subprocess.run(['uname', '-m'], capture_output=True, text=True)
            self.__kernelInfo.machine = result.stdout.strip()
        except subprocess.SubprocessError:
            pass




from .nicInfo import NICInfo
import subprocess


class NICDetectMacOS:
    def __init__(self):
        self.__nicInfo = NICInfo()

    def getNICInfo(self):
        self.__getNICInfo()
        return self.__nicInfo

    def __getNICInfo(self):
        # Placeholder for a more advanced method.
        try:
            interface = subprocess.run(["bash", "-c", "route get default | grep interface"], capture_output=True, text=True).stdout.strip().split(":")[1]
            if_ip = subprocess.run(["bash", "-c", f"ipconfig getifaddr {interface}"], capture_output=True, text=True).stdout.strip()
            self.__nicInfo.nics.append(f"{interface} @ {if_ip}")
            self.__nicInfo.number += 1
        except:
            self.__handleError()

    def __handleError(self):
        self.__nicInfo.nics.append("en0")
        self.__nicInfo.number = 1





import re
import os
import subprocess
from .cpuInfo import CPUInfo
import json


class CPUDetectWindows:
    def __init__(self):
        self.__cpuInfo = CPUInfo()

    def getCPUInfo(self):
        self.__getCPUInfo()
        self.__modelClean()
        if self.__cpuInfo.model == "":
            self.__cpuInfo.model = "Unknown"
        if self.__cpuInfo.model != "":
            self.__cpuInfo.cpu = self.__cpuInfo.model
            if self.__cpuInfo.cores != "":
                self.__cpuInfo.cpu += f" ({self.__cpuInfo.cores})"
            if self.__cpuInfo.frequency != "":
                self.__cpuInfo.cpu += f" @ {self.__cpuInfo.frequency}"
        return self.__cpuInfo

    def __getCPUInfo(self):
        COMMAND = "Get-CimInstance -ClassName Win32_Processor | Select-Object Name,NumberOfLogicalProcessors,MaxClockSpeed | ConvertTo-JSON"

        try:
            result = subprocess.run(["powershell", "-NoProfile", "-Command", COMMAND], capture_output=True, text=True)
        except subprocess.SubprocessError:
            exit(-1)

        res = json.loads(result.stdout)
        name = res["Name"]
        cores = res["NumberOfLogicalProcessors"]
        frequency = round(int(res["MaxClockSpeed"]) / 1000.0, 2)  # GHz
        if "@" in name:
            self.__cpuInfo.model = name.split("@")[0].strip()
        else:
            self.__cpuInfo.model = name
        self.__cpuInfo.cores = str(cores)
        self.__cpuInfo.frequency = f"{frequency:.2f} GHz"

    def __modelClean(self):
        self.__cpuInfo.model = self.__cpuInfo.model.replace("(R)", "")
        self.__cpuInfo.model = self.__cpuInfo.model.replace("(TM)", "")
        self.__cpuInfo.model = self.__cpuInfo.model.replace("CPU ", "")




"""
    In dev.
"""
from ...pyhwUtil import getArch, getDocker
from .hostInfo import HostInfo
import os


class HostDetectLinux:
    def __init__(self):
        self._hostInfo = HostInfo()
        self._arch = getArch()
        self._docker = getDocker()

    def getHostInfo(self):
        self._getModel()
        return self._hostInfo

    def _getModel(self):
        if self._docker:
            self._hostInfo.name = f"General {self._arch} Docker Host"
            self._hostInfo.version = ""
            self._hostInfo.model = self._hostInfo.name + " " + self._hostInfo.version
        else:
            if self._arch in ["x86_64", "x86"]:
                try:
                    with open("/sys/devices/virtual/dmi/id/product_name", "r") as f:
                        product_name = f.read().strip()
                        if product_name.startswith("To be filled by O.E.M."):
                            self._hostInfo.name = f"General {self._arch} Host"
                        else:
                            self._hostInfo.name = product_name
                    with open("/sys/devices/virtual/dmi/id/product_version", "r") as f:
                        version = f.read().strip()
                        if version.startswith("To be filled by O.E.M."):
                            self._hostInfo.version = ""
                        else:
                            self._hostInfo.version = version
                    self._hostInfo.model = self._hostInfo.name + " " + self._hostInfo.version
                except FileNotFoundError:
                    pass
            elif self._arch in ["aarch64", "arm32", "riscv64"]:
                # try to find dmi folder since some arm based desktops and servers may have same structure as x86_64 machines.
                if os.path.exists("/sys/devices/virtual/dmi/id"):
                    try:
                        with open("/sys/devices/virtual/dmi/id/product_name", "r") as f:
                            self._hostInfo.name = f.read().strip()
                        with open("/sys/devices/virtual/dmi/id/product_version", "r") as f:
                            self._hostInfo.version = f.read().strip()
                        self._hostInfo.model = self._hostInfo.name + " " + self._hostInfo.version
                    except FileNotFoundError:
                        pass
                else:
                    # some single board computers may not have dmi folder, try to find model name in device tree.
                    try:
                        with open("/sys/firmware/devicetree/base/model", "r") as f:
                            self._hostInfo.model = f.read().strip()
                    except FileNotFoundError:
                        pass

    def __getHostFamily(self):
        pass

    def __getHostProductName(self):
        pass



from ...pyhwException import OSUnsupportedException


class NICDetect:
    """
    Class for network interface (NIC) detection.
    """
    def __init__(self, os):
        self.OS = os

    def getNICInfo(self):
        """
        Detects the network interfaces (NICs) connected to the system.
        :return: dataclass NICInfo, direct attr: nics
        """
        if self.OS == "linux":
            from .linux import NICDetectLinux
            return NICDetectLinux().getNICInfo()
        elif self.OS == "macos":
            from .macos import NICDetectMacOS
            return NICDetectMacOS().getNICInfo()
        elif self.OS == "freebsd":
            from .bsd import NICDetectBSD
            return NICDetectBSD().getNICInfo()
        elif self.OS == "windows":
            from .windows import NICDetectWindows
            return NICDetectWindows().getNICInfo()
        else:
            raise OSUnsupportedException("Unsupported operating system")



from ...pyhwException import OSUnsupportedException


class CPUDetect:
    def __init__(self, os):
        self.OS = os

    def getCPUInfo(self):
        if self.OS == "linux":
            from .linux import CPUDetectLinux
            return CPUDetectLinux().getCPUInfo()
        elif self.OS == "macos":
            from .macos import CPUDetectMacOS
            return CPUDetectMacOS().getCPUInfo()
        elif self.OS == "freebsd":
            from .bsd import CPUDetectBSD
            return CPUDetectBSD().getCPUInfo()
        elif self.OS == "windows":
            from .windows import CPUDetectWindows
            return CPUDetectWindows().getCPUInfo()
        else:
            raise OSUnsupportedException("Unsupported operating system")



from .linux import MemoryDetectLinux


class MemoryDetectBSD(MemoryDetectLinux):
    def __init__(self):
        MemoryDetectLinux.__init__(self)

    def getMemoryInfo(self):
        self._memoryInfo.memory = f"Pass"
        return self._memoryInfo




from .cpuInfo import CPUInfo
from ...pyhwUtil import sysctlGetString, sysctlGetInt, getArch


class CPUDetectMacOS:
    def __init__(self):
        self.__cpuInfo = CPUInfo()
        self.__arch = getArch()
        self.__pCore = 0
        self.__eCore = 0

    def getCPUInfo(self):
        if self.__arch == "aarch64":
            self.__getCPUModel()
            self.__AppleSiliconBaseFrequency()
            self.__handleAppleSilicon()
            self.__cpuInfo.cpu = f"{self.__cpuInfo.model} ({self.__pCore}P, {self.__eCore}E) @ {self.__cpuInfo.frequency}"
        else:
            self.__getCPUModel()
            self.__getCPUCores()
            self.__getCPUFrequency()
            # need test on Intel Macs.
            self.__cpuInfo.cpu = f"{self.__cpuInfo.model.replace('CPU', f'({self.__cpuInfo.cores})')}"
        return self.__cpuInfo

    def __getCPUModel(self):
        model = sysctlGetString("machdep.cpu.brand_string")
        model = model.replace("(R)", "")
        model = model.replace("(TM)", "")
        self.__cpuInfo.model = model

    def __getCPUCores(self):
        cores = sysctlGetString("hw.logicalcpu_max")
        self.__cpuInfo.cores = cores

    def __getCPUFrequency(self):
        # sysctl doesn't provide the exact CPU frequency on Apple Silicon Macs, there are some indirect methods
        # to get the CPU frequency, but can not integrate with Python directly.
        # C-Based helper module will be added later.
        # See https://github.com/fastfetch-cli/fastfetch/blob/dev/src/detection/cpu/cpu_apple.c for more details.
        freq = sysctlGetString("hw.cpufrequency")
        self.__cpuInfo.frequency = freq

    def __handleAppleSilicon(self):
        nlevels = sysctlGetInt("hw.nperflevels")
        if nlevels is not None and nlevels == 2:    # currently, Apple Silicon chip only has 2 performance levels.
            pcore = sysctlGetInt("hw.perflevel0.logicalcpu_max")    # level 0 is P-core
            ecore = sysctlGetInt("hw.perflevel1.logicalcpu_max")    # level 1 is E-core
            if pcore is not None and ecore is not None:
                self.__pCore = pcore
                self.__eCore = ecore

    def __AppleSiliconBaseFrequency(self):
        # see https://en.wikipedia.org/wiki/Apple_silicon#M_series for more details.
        freq = {
            "Apple M1": "2.30 GHz",
            "Apple M1 Pro": "2.32 GHz",
            "Apple M1 Max": "2.32 GHz",
            "Apple M1 Ultra": "2.32 GHz",
            "Apple M2": "3.50 GHz",
            "Apple M2 Pro": "3.50 GHz",
            "Apple M2 Max": "3.69 GHz",
            "Apple M2 Ultra": "3.70 Ghz",
            "Apple M3": "4.05 GHz",
            "Apple M3 Pro": "4.05 GHz",
            "Apple M3 Max": "4.05 GHz",
            "Apple M4": "4.40 GHz",
            "Apple M4 Pro": "4.40 GHz",
            "Apple M4 Max": "4.40 GHz"
        }
        self.__cpuInfo.frequency = freq.get(self.__cpuInfo.model, "Unknown")



from .uptimeInfo import UptimeInfo
from ...pyhwUtil import sysctlGetString
import re
import subprocess


class UptimeDetectMacOS:
    def __init__(self):
        self._uptimeInfo = UptimeInfo()
        self._boot_time = 0
        self._now = 0

    def getUptimeInfo(self):
        self._getUptime()
        uptime_seconds = self._now - self._boot_time
        hours, remainder = divmod(uptime_seconds, 3600)
        minutes, seconds = divmod(remainder, 60)
        days, hours = divmod(hours, 24)
        days = int(days)
        hours = int(hours)
        minutes = int(minutes)
        seconds = int(seconds)
        if days == 0:
            if hours == 0:
                self._uptimeInfo.uptime = f"{minutes} mins {seconds} secs"
            else:
                self._uptimeInfo.uptime = f"{hours} hours {minutes} mins {seconds} secs"
        else:
            self._uptimeInfo.uptime = f"{days} days {hours} hours {minutes} mins {seconds} secs"
        return self._uptimeInfo

    def _getUptime(self):
        result = sysctlGetString("kern.boottime")
        match = re.search(r'sec = (\d+),', result)
        if match:
            self._boot_time = int(match.group(1))
        else:
            return
        self._now = int(subprocess.check_output(["date", "+%s"]).decode().strip())




"""
 In dev.
"""
import subprocess
from .titleInfo import TitleInfo


class TitleDetectUnix:
    def __init__(self):
        self.__titleInfo = TitleInfo()

    def getTitle(self):
        self.__getTitle()
        return self.__titleInfo

    def __getTitle(self):
        try:
            username_result = subprocess.run(['whoami'], capture_output=True, text=True)
            self.__titleInfo.username = username_result.stdout.strip()
            hostname_result = subprocess.run(['hostname'], capture_output=True, text=True)
            self.__titleInfo.hostname = hostname_result.stdout.strip()
            self.__titleInfo.title = f"{self.__titleInfo.username}@{self.__titleInfo.hostname}"
        except Exception as e:
            pass





from .osInfo import OSInfo
from ...pyhwUtil import getArch
import subprocess


class OSDetectMacOS:
    def __init__(self):
        self.__osInfo = OSInfo()
        self.__ProductName = ""
        self.__ProductVersion = ""
        self.__BuildVersion = ""
        self.__VersionName = ""

    def getOSInfo(self):
        self.__getOS()
        self.__handelOSName()
        if self.__VersionName != "":
            self.__osInfo.prettyName = f"{self.__ProductName} {self.__VersionName} {self.__ProductVersion} {self.__BuildVersion} {getArch()}"
        else:
            self.__osInfo.prettyName = f"{self.__ProductName} {self.__ProductVersion} {self.__BuildVersion} {getArch()}"
        self.__osInfo.id = "macOS"
        return self.__osInfo

    def __getOS(self):
        try:
            result = subprocess.run(["sw_vers"], capture_output=True, text=True)
            sw_vers = result.stdout.split("\n")
        except subprocess.SubprocessError:
            sw_vers = []
        for sw_ver in sw_vers:
            if sw_ver.startswith("ProductName:"):
                self.__ProductName = sw_ver.split(":")[1].strip()
            elif sw_ver.startswith("ProductVersion:"):
                self.__ProductVersion = sw_ver.split(":")[1].strip()
            elif sw_ver.startswith("BuildVersion:"):
                self.__BuildVersion = sw_ver.split(":")[1].strip()

    def __handelOSName(self):
        # Only supports modern macOS
        macOSVersionMap = {
            "11": "Big Sur",
            "12": "Monterey",
            "13": "Ventura",
            "14": "Sonoma",
            "15": "Sequoia"
        }
        if "." in self.__ProductVersion:
            major = self.__ProductVersion.split(".")[0]
        else:
            major = self.__ProductVersion
        version_name = macOSVersionMap.get(major, "")
        if version_name != "":
            self.__VersionName = version_name




from .memoryInfo import MemoryInfo


class MemoryDetectLinux:
    def __init__(self):
        self._memoryInfo = MemoryInfo()

    def getMemoryInfo(self):
        self._getMemory()
        self._memoryInfo.memory = f"{self._memoryInfo.used} MiB / {self._memoryInfo.total} MiB"
        return self._memoryInfo

    def _getMemory(self):
        try:
            with open("/proc/meminfo", "r") as file:
                for line in file:
                    if line.startswith("MemTotal:"):
                        self._memoryInfo.total = round(float(line.split(":")[1].strip()[:-3]) / 1024, 2)
                    elif line.startswith("MemAvailable:"):
                        self._memoryInfo.available = round(float(line.split(":")[1].strip()[:-3]) / 1024, 2)
                self._memoryInfo.used = round(self._memoryInfo.total - self._memoryInfo.available, 2)
        except FileNotFoundError:
            pass




from .pyhwException import OSUnsupportedException, BackendException, GPUNotFoundException, LogoNotFoundException

__all__ = ["OSUnsupportedException", "BackendException", "GPUNotFoundException", "LogoNotFoundException"]



from .osInfo import OSInfo
from ...pyhwException import BackendException
import platform
import winreg


class OSDetectWindows:
    def __init__(self):
        self._osInfo = OSInfo()

    def getOSInfo(self):
        system = platform.system()
        release = platform.release()
        edition = platform.win32_edition()
        machine = platform.machine()
        display = self.__get_windows_version()
        if display != "":
            self._osInfo.prettyName = f"{system} {release} {display} ({edition}) {machine}"
        else:
            self._osInfo.prettyName = f"{system} {release} ({edition}) {machine}"
        if release == "10":
            self._osInfo.id = "windows_10"
        elif release == "11":
            self._osInfo.id = "windows_11"
        else:
            self._osInfo.id = "windows_old"
        return self._osInfo

    @staticmethod
    def __get_windows_version():
        try:
            key = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, r"SOFTWARE\Microsoft\Windows NT\CurrentVersion")
            display_version, _ = winreg.QueryValueEx(key, "DisplayVersion")
            winreg.CloseKey(key)
            return str(display_version)
        except:
            try:
                key = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, r"SOFTWARE\Microsoft\Windows NT\CurrentVersion")
                release_id, _ = winreg.QueryValueEx(key, "ReleaseId")
                winreg.CloseKey(key)
                return str(release_id)
            except:
                return ""



from dataclasses import dataclass


@dataclass
class HostInfo:
    """
    HostInfo class to store host information.
    """
    def __init__(self):
        """
        Initialize HostInfo class, model attribute is required.
        """
        self.os = ""
        self.model = ""
        self.family = ""
        self.name = ""
        self.version = ""
        self.sku = ""
        self.serial = ""
        self.uuid = ""
        self.vendor = ""




from ...pyhwException import OSUnsupportedException


class GPUDetect:
    def __init__(self, os):
        self.OS = os

    def getGPUInfo(self):
        if self.OS == "linux":
            from .linux import GPUDetectLinux
            return GPUDetectLinux().getGPUInfo()
        elif self.OS == "macos":
            from .macos import GPUDetectMacOS
            return GPUDetectMacOS().getGPUInfo()
        elif self.OS == "freebsd":
            from .bsd import GPUDetectBSD
            return GPUDetectBSD().getGPUInfo()
        elif self.OS == "windows":
            from .windows import GPUDetectWindows
            return GPUDetectWindows().getGPUInfo()
        else:
            raise OSUnsupportedException("Unsupported operating system")



from dataclasses import dataclass


@dataclass
class CPUInfo:
    """
    Class to store CPU information. Print key: cpu
    """
    def __init__(self):
        self.cpu = ""
        self.model = ""
        self.cores = ""
        self.frequency = ""
